'
'DEGIF6.BAS - No frills GIF decoder for the VGA's 320x200x256 mode.
'By Rich Geldreich 1993 (Public domain, use as you wish.)
'This version should properly decode all LZW encoded images in
'GIF image files. I've finally added GIF89a and local colormap
'support, so it more closely follows the GIF specification. It
'still doesn't support the entire GIF89a specification, but it'll
'show most GIF files fine.
'The GIF decoding speed of this program isn't great, but I'd say
'for an all QB/PDS decoder it's not bad!
'Note: This program does not stop decoding the GIF image after the
'rest of the scanlines become invisible! This happens with images
'larger than the 320x200 screen. So if the program seems to be
'just sitting there, accessing your hard disk, don't worry...
'It'll beep when it's done.
'Modified 04/10/99 by Fred Buffington (Buff) for 16 color Gifs
'will print any 16 of 256K colors defined.
DEFINT A-Z
'Prefix() and Suffix() hold the LZW phrase dictionary.
'OutStack() is used as a decoding stack.
'ShiftOut() as a power of two table used to quickly retrieve the LZW
'multibit codes.
DIM Prefix(4095), Suffix(4095), outstack(4095), ShiftOut(8)
 
'The following line is for the QB environment(slow).
DIM ybase AS LONG, Powersof2(11) AS LONG, WorkCode AS LONG
'For a little more speed, unremark the next line and remark the one
'above, before you compile... You'll get an overflow error if the
'following line is used in the QB environment, so change it back.
'DIM ybase AS INTEGER, Powersof2(11) AS INTEGER, WorkCode AS INTEGER
COMMON finam$
COMMON line.no%
ON ERROR GOTO error.routine

10
'Precalculate power of two tables for fast shifts.
'FOR FL% = 2 TO 3
    FOR a = 0 TO 8: ShiftOut(8 - a) = 2 ^ a: NEXT
    FOR a = 0 TO 11: Powersof2(a) = 2 ^ a: NEXT
    FIRSTTIME = 0
'Get GIF filename.

    a$ = COMMAND$
    IF a$ = "" THEN
        SCREEN 2: SCREEN 0
        LOCATE 24, 1: PRINT "Enter Path and file name for GIF file: ";
        LINE INPUT "", a$: IF a$ = "" THEN END
    END IF
'Add GIF extension if the given filename doesn't have one.
FOR a = LEN(a$) TO 1 STEP -1
    SELECT CASE MID$(a$, a, 1)
    CASE "\", ":": EXIT FOR
    CASE ".": Extension = -1: EXIT FOR
    END SELECT
NEXT
IF Extension = 0 THEN a$ = a$ + ".GIF"
 
'Open file for input so QB stops with an error if it doesn't exist.
CLOSE #1: OPEN a$ FOR INPUT AS #1: CLOSE #1
OPEN a$ FOR BINARY AS #1
 
'Check to see if GIF file. Ignore GIF version number.
a$ = "      ": GET #1, , a$
IF LEFT$(a$, 3) <> "GIF" THEN PRINT "Not a GIF file.": END
 
'Get logical screen's X and Y resolution.
GET #1, , TotalX: GET #1, , TotalY: GOSUB GetByte
'Calculate number of colors and find out if a global palette exists.
numcolors = 2 ^ ((a AND 7) + 1): NoPalette = (a AND 128) = 0
'Retrieve background color.
GOSUB GetByte: Background = a
 
'Get aspect ratio and ignore it.
GOSUB GetByte
 
'Retrieve global palette if it exists.
IF NoPalette = 0 THEN p$ = SPACE$(numcolors * 3): GET #1, , p$
 
DO 'Image decode loop
 
'Skip by any GIF extensions.
'(With a few modifications this code could also fetch comments.)
DO
    'Skip by any zeros at end of image (why must I do this? the
    'GIF spec never mentioned it)
    DO
        IF EOF(1) THEN EXIT DO 'if at end of file, exit
        GOSUB GetByte
    LOOP WHILE a = 0           'loop while byte fetched is zero
 
    SELECT CASE a
    CASE 44  'We've found an image descriptor!
        EXIT DO
    CASE 59  'GIF trailer, stop decoding.
        GOTO AllDone
    CASE IS <> 33
        'PRINT "Unknown GIF extension type.": END
        EXIT DO
    END SELECT
'    'Skip by blocked extension data.
    
    GOSUB GetByte
    DO: GOSUB GetByte: a$ = SPACE$(a): GET #1, , a$: LOOP UNTIL a = 0
LOOP
'Get image's start coordinates and size.
GET #1, , XStart: GET #1, , YStart: GET #1, , Xlength: GET #1, , Ylength
xend = XStart + Xlength: YEnd = YStart + Ylength
 
'Check for local colormap, and fetch it if it exists.
GOSUB GetByte
IF (a AND 128) THEN
    NoPalette = 0
    numcolors = 2 ^ ((a AND 7) + 1)
    p$ = SPACE$(numcolors * 3): GET #1, , p$
END IF
 
'Check for interlaced image.
Interlaced = (a AND 64) > 0: PassNumber = 0: PassStep = 8
 
'Get LZW starting code size.
GOSUB GetByte
 
'Calculate clear code, end of stream code, and first free LZW code.
ClearCode = 2 ^ a
EOSCode = ClearCode + 1
FirstCode = ClearCode + 2: NextCode = FirstCode
StartCodeSize = a + 1: CodeSize = StartCodeSize
 
'Find maximum code for the current code size.
StartMaxCode = 2 ^ (a + 1) - 1: MaxCode = StartMaxCode
 
BitsIn = 0: BlockSize = 0: BlockPointer = 1
 
x = XStart: y = YStart: ybase = y * 320&

'Set screen 13 in not set yet.
IF FIRSTTIME = 0 THEN
    'Go to VGA mode 13 (320x200x256).
    IF numcolors > 16 THEN
        SCREEN 13: DEF SEG = &HA000
    ELSE
        SCREEN 12: DEF SEG = &HA000
    END IF
END IF
 
'Set palette, if there was one.
IF NoPalette = 0 THEN
    'Use OUTs for speed.
    IF numcolors > 16 THEN
        OUT &H3C8, 0
        FOR a = 1 TO numcolors * 3: OUT &H3C9, ASC(MID$(p$, a, 1)) \ 4: NEXT
        'Save palette of image to disk.
        'OPEN "pal." FOR BINARY AS #2: PUT #2, , P$: CLOSE #2
    ELSE
        ans$ = "Y"
        IF UCASE$(ans$) = "Y" THEN
            FOR k = 0 TO numcolors - 1
                PALETTE k, 0
'DO  'loop here and update the palette and data w/ each key hit
                red% = ASC(MID$(p$, k * 3 + 1, 1)) / 5
                green% = ASC(MID$(p$, k * 3 + 2, 1)) / 5
                blue% = ASC(MID$(p$, k * 3 + 3, 1)) / 5
                a& = (65536 * blue%) + (256 * green%) + red%   'calc the new palette
                PALETTE k, a&   'display the new palette
            NEXT k
        END IF
    END IF

END IF
 
IF FIRSTTIME = 0 THEN
  'Clear entire screen to background color. This isn't
  'done until the image's palette is set, to avoid flicker
  'on some GIFs.
    LINE (0, 0)-(319, 199), Background, BF
    FIRSTTIME = -1
END IF
 
'Decode LZW data stream to screen.
DO
    'Retrieve one LZW code.
    GOSUB GetCode
    'Is it an end of stream code?
    IF Code <> EOSCode THEN
       
        'Is it a clear code? (The clear code resets the sliding
        'dictionary - it *should* be the first LZW code present
        ' in the data stream.)
 
        IF Code = ClearCode THEN
            NextCode = FirstCode
            CodeSize = StartCodeSize
            MaxCode = StartMaxCode
            GOSUB GetCode
            CurCode = Code: LastCode = Code: lastpixel = Code
            IF (TotalX < 325 AND TotalY < 205) OR numcolors <= 16 THEN
                IF numcolors <= 16 THEN
                clr1% = lastpixel
            IF UCASE$(ans$) = "N" THEN
                IF lastpixel MOD 16 = 9 THEN
                    clr1% = 12
                'ELSEIF lastpixel MOD 16 = 1 THEN
                '    clr1% = 4
                ''?ELSEIF lastpixel MOD 16 = 4 THEN
                ''?    clr1% = 1
                ELSEIF lastpixel MOD 16 = 12 THEN
                    clr1% = 9
            
                ELSEIF lastpixel MOD 16 = 11 THEN
                    clr1% = 14
'                ELSEIF lastpixel mod 16= 3 THEN
'                    clr1% = 6
            '''    clr1% = lastpicel MOD 16
                END IF
            END IF
                PSET (x, y), clr1%
                   'PSET (X, Y), colr
                ELSE
                    IF lastpixel <> 0 THEN
                        POKE x + ybase, lastpixel
                    END IF
                END IF
            END IF
          IF numcolors > 16 THEN
            IF (TotalX >= 325 AND TotalX <= 640) THEN  '''OR (TotalY > 205 AND TotalY < 481) and numcolors>16 THEN
              IF count / 2 = INT(count / 2) AND count < 800 THEN
                IF lastpixel <> 0 THEN
                    IF x < 641 AND y < 401 THEN PSET (x / 2, y / 2), lastpixel''POKE x / 4 + ybase, LastPixel
                END IF
               END IF
            END IF
            IF TotalX > 640 OR numcolors > 16 THEN '''OR TotalY > 480 THEN
              IF count / 4 = INT(count / 4) AND count < 1200 THEN
                IF lastpixel <> 0 THEN
                    IF x < 1280 AND y < 1200 THEN PSET (x / 4, y / 4), lastpixel''POKE x / 4 + ybase, LastPixel
                END IF
               END IF
            END IF
          END IF
''this moves picture to left            IF x > 120 AND x < 440 AND y < 200 THEN POKE x + ybase - 120, LastPixel
            x = x + 1
            IF x = xend THEN
                    count = count + 1
                    GOSUB NextScanLine
                
             END IF
        ELSE
            CurCode = Code: StackPointer = 0
 
            'Have we entered this code into the dictionary yet?
            IF Code >= NextCode THEN
                IF Code > NextCode THEN GOTO AllDone 'Bad GIF if
                ' this happens.
                'mimick last code if we haven't entered the
                ' requested
                'code into the dictionary yet
                CurCode = LastCode
                outstack(StackPointer) = lastpixel
                StackPointer = StackPointer + 1
            END IF
 
            'Recursively get each character of the string.
            'Since we get the characters in reverse, "push" them
            'onto a stack so we can "pop" them off later.
            'Hint: There is another, much faster way to
            ' accomplish
            'this that doesn't involve a decoding stack at all...
            DO WHILE CurCode >= FirstCode
                outstack(StackPointer) = Suffix(CurCode)
                StackPointer = StackPointer + 1
                CurCode = Prefix(CurCode)
            LOOP
 
            lastpixel = CurCode
            IF (TotalX < 325 AND TotalY < 205) OR numcolors <= 16 THEN
                 IF numcolors <= 16 THEN
                clr1% = lastpixel
             IF UCASE$(ans$) = "N" THEN
                IF lastpixel = 9 THEN
                    clr1% = 12
                'ELSEIF lastpixel MOD 16 = 1 THEN
                '    clr1% = 4
                ''?ELSEIF lastpixel MOD 16 = 4 THEN
               ''?     clr1% = 1
                ELSEIF lastpixel MOD 16 = 12 THEN
                    clr1% = 9
             
                ELSEIF lastpixel MOD 16 = 11 THEN
                    clr1% = 14
'                ELSEIF lastpixel mod 16= 3 THEN
'                    clr1% = 6
             '''   clr1% = lastpixel MOD 16
                END IF
             END IF
                    PSET (x, y), clr1%
                     
                      'PSET (X, Y), colr
                 ELSE
                    POKE x + ybase, lastpixel
                 END IF
            END IF
          IF numcolors > 16 THEN
            IF (TotalX >= 325 AND TotalX <= 640) THEN '''OR (TotalY > 205 AND TotalY < 481) THEN
                IF count / 2 = INT(count / 2) AND count < 800 THEN
                    IF lastpixel <> 0 THEN
                        IF x < 640 AND y < 401 THEN PSET (x / 2, y / 2), lastpixel''POKE x / 2 + ybase, lastPixel
                    END IF
                END IF
             END IF
            IF TotalX > 640 THEN '''OR TotalY > 480 THEN
                IF count / 4 = INT(count / 4) AND count < 1200 THEN
                    IF lastpixel <> 0 THEN
                        IF x < 1280 AND y < 1200 THEN PSET (x / 4, y / 4), lastpixel''POKE x / 2 + ybase, lastPixel
                    END IF
                END IF
             END IF
          END IF
''this moves picture to left            IF x > 120 AND x < 440 AND y < 200 THEN POKE x + ybase - 120, LastPixel
            x = x + 1
            IF y > 800 THEN
                CLOSE
                x$ = INKEY$: IF x$ <> "" THEN SYSTEM
                SLEEP 3
                'CHAIN "Degif6"''x = xend
            END IF
            IF x = xend THEN
                count = count + 1
                GOSUB NextScanLine
            END IF
           
            '"Pop" each character onto the display.
            FOR a = StackPointer - 1 TO 0 STEP -1
                IF (TotalX < 325 AND TotalY < 205) OR numcolors <= 16 THEN
                     IF numcolors <= 16 THEN
                clr1% = outstack(a)
             IF UCASE$(ans$) = "N" THEN
                IF outstack(a) MOD 16 = 9 THEN
                    clr1% = 12
                'ELSEIF outstack(a) MOD 16 = 1 THEN
                '    clr1% = 4
                ''?ELSEIF outstack(a) MOD 16 = 4 THEN
                ''?    clr1% = 1
                ELSEIF outstack(a) MOD 16 = 12 THEN
                    clr1% = 9
             
                ELSEIF outstack(a) MOD 16 = 11 THEN
                    clr1% = 14
'                ELSEIF outstack(a) = 3 THEN
'                    clr1% = 6
           '''     clr1% = outstack(a) MOD 16
                END IF
            END IF
                   PSET (x, y), clr1%
                        
                         'PSET (X, Y), colr
                     ELSE
                        POKE x + ybase, outstack(a)
                     END IF
                END IF
              IF numcolors > 16 THEN
                IF (TotalX >= 325 AND TotalX <= 640) THEN '''OR (TotalY > 205 AND TotalY < 481) THEN
                    IF count / 2 = INT(count / 2) AND count < 800 THEN
                       IF outstack(a) <> 0 THEN
                        IF x < 641 AND y < 401 THEN PSET (x / 2, y / 2), outstack(a)
                       END IF
                    END IF
               END IF
                IF TotalX > 640 THEN '''OR TotalY > 480 and numcolors>16 THEN
                    IF count / 4 = INT(count / 4) AND count < 1200 THEN
                       IF outstack(a) <> 0 THEN
                        IF x < 1280 AND y < 1200 THEN PSET (x / 4, y / 4), outstack(a)
                       END IF
                    END IF
               END IF
              END IF
''this moves picture to left                IF x > 120 AND x < 440 AND y < 205 THEN POKE x + ybase - 120, OutStack(A)
                x = x + 1
                IF y > 800 THEN
                    CLOSE
                    x$ = INKEY$: IF x$ <> "" THEN SYSTEM
                    SLEEP 3
                    'CHAIN "Degif6"''x = xend
                END IF
                IF x = xend THEN
                    count = count + 1
                    GOSUB NextScanLine
                END IF
           
            NEXT
 
            'Can we put this new string into our dictionary?
            ' (Some GIF
            'encoders will wait a bit when the dictionary is full
            'before sending a clear code- this increases
            ' compression
            'because the dictionary's contents are thrown away
            ' less
            'often.)
            IF NextCode < 4096 THEN
                'Store new string in the dictionary for later
                Prefix(NextCode) = LastCode
                Suffix(NextCode) = lastpixel
                NextCode = NextCode + 1
            'Time to increase the LZW code size?
                IF (NextCode > MaxCode) AND (CodeSize < 12) THEN
                    CodeSize = CodeSize + 1
                    MaxCode = MaxCode * 2 + 1
                END IF
            END IF
            LastCode = Code
        END IF
    END IF
LOOP UNTIL Code = EOSCode
 
LOOP
 
AllDone:
 
'Save image and palette to BSAVE file.
'DEF SEG = &HA000
'OUT &H3C7, 0
'FOR a = 0 TO 767
'    POKE a + 64000, INP(&H9)
'NEXT
 
'BSAVE "pic.bas", 0, 64768
 
'Load images saved with the above code with this:
'DEF SEG= &HA000
'BLOAD "Pic.Bas"
'OUT &H3C8, 0
'FOR a = 0 To 767
'     OUT &H3C9, Peek(a+ 64000)
'NEXT
 
'BEEP: DO: LOOP WHILE INKEY$ <> "": DO: LOOP UNTIL INKEY$ <> ""

'END
CHECK.IT = -1
START.TIME = VAL(MID$(TIME$, 4, 2)) * 60 + VAL(MID$(TIME$, 7, 2))

x$ = INKEY$
IF x$ = CHR$(27) THEN CLOSE : DEF SEG : END
'BSAVE image$, 0, 64768
'IF finam$ = fil$(6) THEN CLOSE : DEF SEG : STOP
'SLEEP 3
WHILE INKEY$ = "": WEND
IF numcolors <= 16 AND ans$ = "Y" THEN
   'reset palette
   PALETTE
   SCREEN 0

END IF
END
'CHAIN "DEGIF6"
'Slowly reads one byte from the GIF file...
GetByte: a$ = " ": GET #1, , a$: a = ASC(a$): RETURN
 
'Moves down one scanline. If the GIF is interlaced, then the number
'of scanlines skipped is based on the current pass.
NextScanLine:
    IF Interlaced THEN
        y = y + PassStep
        IF y >= YEnd THEN
            PassNumber = PassNumber + 1
            SELECT CASE PassNumber
            CASE 1: y = 4: PassStep = 8
            CASE 2: y = 2: PassStep = 4
            CASE 3: y = 1: PassStep = 2
            END SELECT
        END IF
    ELSE
        y = y + 1
    END IF
    x = XStart: ybase = y * 320&
RETURN
 
'Reads a multibit code from the data stream. Look ma, see how
'simple it is now!
GetCode:
    WorkCode = LastChar \ ShiftOut(BitsIn)
  'Loop while more bits are needed.
    DO WHILE CodeSize > BitsIn
'Reads a byte from the LZW data stream. Since the data stream is
'blocked, a check is performed for the end of the current block
'before each byte is fetched.
        IF BlockPointer > BlockSize THEN
      'Retrieve block's length
            GOSUB GetByte: BlockSize = a
            a$ = SPACE$(BlockSize): GET #1, , a$
            BlockPointer = 1
        END IF
     'Yuck, ASC() and MID$() aren't that fast.
        LastChar = ASC(MID$(a$, BlockPointer, 1))
        BlockPointer = BlockPointer + 1
     'Append 8 more bits to the input buffer
        WorkCode = WorkCode OR LastChar * Powersof2(BitsIn)
        BitsIn = BitsIn + 8
    LOOP
  'Take away x number of bits.
    BitsIn = BitsIn - CodeSize
  'Return code to caller.
    Code = WorkCode AND MaxCode
RETURN
 

error.routine:
    SELECT CASE ERR
        CASE 5
            CLOSE
            'line.no% = 0
            SLEEP 3
            END
            CHAIN "degif6"
           
            RESUME NEXT
       
        CASE 62
            CLOSE
            line.no% = 0
            SLEEP 3
            END
            CHAIN "degif6"
            STOP
        CASE ELSE
            RESUME NEXT
         END SELECT
     CLOSE


